package ch.unige.solidify.auth.oidc;

import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.core.env.Environment;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import ch.unige.solidify.auth.config.SsoAppConfig;
import ch.unige.solidify.auth.util.AuthUtility;
import ch.unige.solidify.auth.util.OIDCUtil;
import ch.unige.solidify.auth.util.OidcTokenExpiredException;
import ch.unige.solidify.model.sso.SsoUser;

@Component
public class OidcUserFactory {

  public static final String OIDC_ACCESS_TOKEN = "oidc_access_token";
  public static final String OIDC_ACCESS_TOKEN_EXPIRES = "oidc_access_token_expires";
  public static final String OIDC_USER_CACHE_NAME = "cache_for_oidc";
  private static final Logger log = LoggerFactory.getLogger(OidcUserFactory.class);
  private static final int TEN_MINUTES = 600000;

  @Autowired
  Environment environment;
  private SsoAppConfig ssoAppConfig;
  private OidcUserInfoService oidcUserInfoService;

  @Autowired
  public OidcUserFactory(SsoAppConfig ssoAppConfig, OidcUserInfoService oidcUserInfoService) {
    this.ssoAppConfig = ssoAppConfig;
    this.oidcUserInfoService = oidcUserInfoService;
  }

  @Cacheable(value = OIDC_USER_CACHE_NAME, key = "{#p0, #p1}")
  public SsoUser get(String accesstoken, long tokenExpiresTime, Map<String, String> headers)
          throws OidcTokenExpiredException {
    if (log.isDebugEnabled()) {
      log.debug(
              "OidcUserFactory.get(" + accesstoken + ", " + tokenExpiresTime + ", " + headers + ")");
    }
    OIDCUtil.generateOidcTokenExpiredExceptionIfTokenNeedFresh(accesstoken, tokenExpiresTime);
    OIDCUser result = oidcUserInfoService.getOidcUser(accesstoken, tokenExpiresTime);
    result.setAllProperties(headers);
    result.setUserContext(AuthUtility.calculeUserContext(headers, ssoAppConfig));
    return OIDCUser.toUserData(result);
  }

  @CacheEvict(allEntries = true, cacheNames = { OIDC_USER_CACHE_NAME })
  @Scheduled(fixedDelay = TEN_MINUTES)
  public void reFreshOidcUserCache() {
  }

}
