package ch.unige.solidify.auth.shibboleth;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.xml.bind.annotation.XmlRootElement;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.env.Environment;
import org.springframework.stereotype.Component;

import ch.unige.solidify.auth.config.SsoAppConfig;
import ch.unige.solidify.auth.util.AuthUtility;
import ch.unige.solidify.auth.util.SsoUserData;
import ch.unige.solidify.model.sso.SsoUser;
import ch.unige.solidify.util.StringTool;

/**
 * Read user's attributes from HTTP headers and returns a SsoUser object from them.
 *
 * @link https://aai-viewer.switch.ch/aai/
 */
@Component
public class ShibbolethUserFactory {

  private static final Logger log = LoggerFactory.getLogger(ShibbolethUserFactory.class);

  private static final String SEPARATOR = ";";

  private ShibbolethHeaderConfig config;
  private SsoAppConfig ssoAppConfig;
  private Environment environment;

  @Autowired
  public ShibbolethUserFactory(ShibbolethHeaderConfig config, SsoAppConfig ssoAppConfig,
          Environment environment) {
    this.config = config;
    this.ssoAppConfig = ssoAppConfig;
    this.environment = environment;

  }

  public SsoUser get(Map<String, String> attributes) {
    return new Reader(attributes, config, ssoAppConfig, environment);
  }

  /**
   * Read AAI attributes from the a (Http Header) map
   */
  @XmlRootElement
  public static class Reader extends SsoUserData {
    private static final long serialVersionUID = -6336989404081508577L;

    private Map<String, String> attributes;
    private ShibbolethHeaderConfig config;
    private SsoAppConfig ssoAppConfig;
    private Environment environment;

    Reader(final Map<String, String> attributes, ShibbolethHeaderConfig config,
            SsoAppConfig ssoAppConfig, Environment environment) {
      this.attributes = attributes;
      this.config = config;
      this.ssoAppConfig = ssoAppConfig;
      this.environment = environment;
      setUserData();
    }

    @Override
    public boolean isEmpty() {
      return StringTool.isNullOrEmpty(getUniqueId());
    }

    private void setUserData() {
      super.setAllProperties(this.attributes);
      super.setUniqueid(calculeUniqueid());
      super.setFirstname(calculeFirstname());
      super.setLastname(calculeLastname());
      super.setEmail(calculeEmail());
      super.setHomeorganization(calculeHomeorganization());
      super.setEmployeeType(calculeEmployeeType());
      super.setUnigeChStudentCategory(calculeUnigeChStudentCategory());
      super.setIsmemberOf(calculeIsmemberOf());
      super.setHomeorganizationtype(calculeHomeorganizationtype());
      super.setAffiliation(calculeAffiliation());
      super.setUnigechoucode(calculeUnigechoucode());
      super.setUnigechstudentoucode(calculeUnigechstudentoucode());
      super.setUnigechemployeeoucode(calculeUnigechemployeeoucode());
      super.setOrgunitdn(calculeOrgunitdn());
      super.setMatriculationnumber(calculeMatriculationnumber());
      super.setCnindividu(calculeCnindividu());
      super.setName(calculeName());
      super.setUserContext(AuthUtility.calculeUserContext(attributes, ssoAppConfig));
      super.setRoles(calculeApplicationRoles());
    }

    /**
     * The fist item in the attribute array for the specified key
     *
     * @param key The key to search for
     * @return The first element of the array matching the key.
     */
    private String first(final String key) {
      if (!attributes.containsKey(key)) {
        return "";
      }
      String result = attributes.get(key);
      result = StringTool.isNullOrEmpty(result) ? "" : result;
      return result;
    }

    private String calculeUniqueid() {
      String key = config.getUniqueId();
      String uniqueId = first(key);
      if (uniqueId != null && !uniqueId.isEmpty()) {
        return uniqueId;
      }
      if (AuthUtility.isProfilesActive(environment, "dev", "qa")) {
        uniqueId = attributes.get(SsoUser.CONNECTED_USER_UNIQUE_ID);
      }

      return uniqueId;
    }

    private String calculeFirstname() {
      String key = config.getFirstName();
      return first(key);
    }

    private String calculeLastname() {
      String key = config.getLastName();
      return first(key);
    }

    private String calculeEmail() {
      String key = config.getEmail();
      return first(key);
    }

    private String calculeHomeorganization() {
      String key = config.getOrganization();
      return first(key);
    }

    private Set<String> calculeEmployeeType() {
      String key = config.getEmployeeType();
      if (StringTool.isNullOrEmpty(first(key))) {
        return Collections.synchronizedSet(new HashSet<>());
      }
      String[] items = first(key).split(SEPARATOR);
      return Collections.synchronizedSet(new HashSet<>(Arrays.asList(items)));
    }

    private Set<String> calculeUnigeChStudentCategory() {
      String key = config.getUnigeChStudentCategory();
      if (StringTool.isNullOrEmpty(first(key))) {
        return Collections.synchronizedSet(new HashSet<>());
      }
      String[] items = first(key).split(SEPARATOR);
      return Collections.synchronizedSet(new HashSet<>(Arrays.asList(items)));
    }

    private Set<String> calculeIsmemberOf() {
      String key = config.getIsMemberOf();
      String value = first(key);
      if (StringTool.isNullOrEmpty(value)) {
        return Collections.synchronizedSet(new HashSet<>());
      }
      String[] items = value.split(SEPARATOR);
      return Collections.synchronizedSet(new HashSet<>(Arrays.asList(items)));
    }

    private String calculeHomeorganizationtype() {
      String key = config.getOrganizationType();
      return first(key);
    }

    private Set<String> calculeAffiliation() {
      String key = config.getAffiliation();
      String[] items = first(key).split(SEPARATOR);
      return Collections.synchronizedSet(new HashSet<>(Arrays.asList(items)));
    }

    private String calculeUnigechoucode() {
      String key = config.getUnigeChOuCode();
      return first(key);
    }

    private String calculeUnigechstudentoucode() {
      String key = config.getUnigeChStudentOuCode();
      return first(key);
    }

    private String calculeUnigechemployeeoucode() {
      String key = config.getUnigeChEmployeeOuCode();
      return first(key);
    }

    private String calculeOrgunitdn() {
      String key = config.getOrgunitdn();
      return first(key);
    }

    private String calculeMatriculationnumber() {
      String key = config.getMatriculationnumber();
      return first(key);
    }

    private String calculeCnindividu() {
      String key = config.getUniqueId();
      String value = first(key);
      int index = value.indexOf("@unige.ch");
      if (value.indexOf("@unige.ch") > 0) {
        return value.substring(0, index);
      }
      return null;
    }

    private String calculeName() {
      String firstname = getFirstName();
      String lastname = getLastName();

      if (StringTool.isNullOrEmpty(firstname) && StringTool.isNullOrEmpty(lastname)) {
        return "";
      }

      if (StringTool.isNullOrEmpty(firstname)) {
        return lastname;
      }

      if (StringTool.isNullOrEmpty(lastname)) {
        return firstname;
      }

      return firstname + " " + lastname;
    }

    @SuppressWarnings("unchecked")
    private void setRoleWithcalculeResult(Object foundValues, Set<String> waitedValues,
            String applicationRoleName, Set<String> rolesList) {

      if (foundValues instanceof Set) {
        Set<String> foundSetValue = (Set<String>) foundValues;
        foundSetValue.retainAll(waitedValues);
        if (!foundSetValue.isEmpty()) {
          rolesList.add(applicationRoleName);
        }
        return;
      }

      if (foundValues instanceof String) {// if method returns a String value
        String foundStringValue = (String) foundValues;
        if (waitedValues.contains(foundStringValue)) {
          rolesList.add(applicationRoleName);
        }
      }
    }

    private Set<String> calculeApplicationRoles() {
      Set<String> roleList = Collections.synchronizedSet(new HashSet<>());
      List<SsoAppConfig.Group> applicationGroups = ssoAppConfig.getGroups();

      for (SsoAppConfig.Group group : applicationGroups) {
        if (group.isValid() || (!group.getOrg().equalsIgnoreCase(getHomeOrganization()))) {
          break;
        }
        Method[] allMethods = super.getClass().getDeclaredMethods();
        for (Method m : allMethods) {
          if (m.getName().equalsIgnoreCase("calcule" + group.getCriteria())) {
            try {
              Object mothodeResult = m.invoke(this);
              Set<String> waitedValues = new HashSet<>(group.getCriteriaValues());
              setRoleWithcalculeResult(mothodeResult, waitedValues, group.getRole(), roleList);
            } catch (Exception e) {
              log.error("Application Role and grouper mapping error", e);
              break;
            }
          }
        }
      }
      return roleList;
    }

  }
}
