package ch.unige.solidify.test.auth;

import ch.unige.solidify.auth.SsoAuthentication;
import ch.unige.solidify.model.sso.SsoUser;
import ch.unige.solidify.auth.SsoUserDetails;
import ch.unige.solidify.auth.config.SsoAppConfig;
import ch.unige.solidify.auth.shibboleth.ShibbolethHeaderConfig;
import ch.unige.solidify.auth.shibboleth.ShibbolethUserFactory;
import ch.unige.solidify.auth.util.SsoUserData;
import ch.unige.solidify.test.auth.config.MockApplication;
import ch.unige.solidify.test.auth.data.MockData;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.context.TestConfiguration;
import org.springframework.core.env.Environment;
import org.springframework.test.context.junit4.SpringRunner;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static org.junit.Assert.*;

/**
 * unit test for Creation of a user from HashMap who contains shibboleth headers value Creation of a
 * user from HashMap who contains OID headers value class SsoUserdetails and SsoAuthentication
 */

@RunWith(SpringRunner.class)
@TestConfiguration
@SpringBootTest(classes = MockApplication.class)
public class ShibbolethSsoUserTest {

  @Autowired
  private ShibbolethHeaderConfig config;
  @Autowired
  private SsoAppConfig ssoRoleConfig;
  @Autowired
  private Environment environment;
  private Map<String, Object> oidcHeaders = null;
  private Map<String, String> shibboHeaders = null;

  @Before
  public void setup() {
    oidcHeaders = MockData.getOidcHeadersMockData();
    for (String key : oidcHeaders.keySet()) {
      oidcHeaders.put(key, oidcHeaders.get(key));
    }

    shibboHeaders = MockData.getShibboHeadersMockData();
    for (String key : shibboHeaders.keySet()) {
      shibboHeaders.put(key, shibboHeaders.get(key));
    }
  }

  @Test
  public void testConfig() {
    List<SsoAppConfig.Group> groups = ssoRoleConfig.getGroups();
    for (SsoAppConfig.Group group : groups) {
      String value = group.getRole();
      group.setRole(null);
      assertTrue(group.isValid());
      group.setRole(value);

      List<String> values = group.getCriteriaValues();
      group.setCriteriaValues(null);
      assertTrue(group.isValid());

      group.setCriteriaValues(new ArrayList<>());
      assertTrue(group.isValid());
      group.setCriteriaValues(values);

    }
  }

  @Test
  public void testShibboUser() {
    ShibbolethUserFactory userFactory = new ShibbolethUserFactory(config, ssoRoleConfig, environment);
    SsoUser ssoUser = userFactory.get(shibboHeaders);
    assertTrue(ssoUser.toString().length() > 10);
    assertEquals(MockData.ShibboSystemId, ssoUser.getSystemId());
    assertEquals(MockData.employeeNumber, ssoUser.getCnindividu());

    Set<String> roles = ssoUser.getRoles();
    assertEquals(MockData.ShibboUniqueId, ssoUser.getUniqueId());
    assertTrue(roles.contains("techAdmin"));
    assertTrue(roles.contains("editor"));
    Map<String, String> allProperties = ssoUser.getAllProperties();
    assertNotNull(allProperties);
    shibboHeaders.put("uid", "test");
    assertEquals("test", ssoUser.getAllProperties().get("uid"));
  }

  @Test(expected = IllegalStateException.class)
  public void testSsoUserPassword() {
    SsoUserDetails details = new SsoUserDetails();
    assertNotNull(details.getPassword());
  }

  @Test
  public void testWrongValue() {
    ShibbolethUserFactory userFactory = new ShibbolethUserFactory(config, ssoRoleConfig, environment);
    // unique id wrong format
    String value = shibboHeaders.get("uniqueId");
    shibboHeaders.put("uniqueid", "test@unige.ch");
    SsoUser ssoUser = userFactory.get(shibboHeaders);
    shibboHeaders.put("uniqueid", value);
    assertEquals(0, ssoUser.getSystemId());

    String givenname = shibboHeaders.get("givenname");
    String lastname = shibboHeaders.get("surname");

    // without surname
    shibboHeaders.put("surname", null);
    ssoUser = userFactory.get(shibboHeaders);
    shibboHeaders.put("surname", lastname);
    assertEquals(givenname, ssoUser.getName());

    // without first
    shibboHeaders.put("givenname", null);
    ssoUser = userFactory.get(shibboHeaders);
    shibboHeaders.put("givenname", givenname);
    assertEquals(lastname, ssoUser.getName());

    // without first and laist
    shibboHeaders.put("givenname", null);
    shibboHeaders.put("surname", null);
    ssoUser = userFactory.get(shibboHeaders);
    shibboHeaders.put("givenname", givenname);
    shibboHeaders.put("surname", lastname);
    assertEquals("", ssoUser.getName());

  }

  @Test
  public void testSsoUserDetail() {
    SsoUserDetails details = new SsoUserDetails();
    assertTrue(details.isAccountNonExpired());
    assertTrue(details.isAccountNonLocked());
    assertTrue(details.isCredentialsNonExpired());
    assertTrue(details.isEnabled());

    SsoUserData data = new SsoUserData();
    data.setUniqueid("3123@test");
    SsoUserDetails details2 = new SsoUserDetails(data);
    assertTrue(!details2.isEmpty());
    assertNotNull(details2.getUsername());
  }

  @Test(expected = IllegalStateException.class)
  public void testAuthenticationCredential() {
    SsoUserDetails details = new SsoUserDetails();
    SsoAuthentication aut = new SsoAuthentication(details);
    assertNotNull(aut.getCredentials());
  }

  @Test(expected = IllegalStateException.class)
  public void testSetAuthenticatedTrue() {
    SsoUserDetails details = new SsoUserDetails();
    SsoAuthentication aut = new SsoAuthentication(details);
    aut.setAuthenticated(true);
  }

  @Test
  public void testSetAuthenticatedFalse() {
    SsoUserDetails details = new SsoUserDetails();
    SsoAuthentication aut = new SsoAuthentication(details);
    aut.setAuthenticated(false);
    assertNull(aut.getDetails());
  }

}
