package ch.unige.solidify.controller;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;
import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.methodOn;

import java.util.ArrayList;
import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.SubResourceContainer;
import ch.unige.solidify.rest.Tool;
import ch.unige.solidify.service.CompositeResourceService;

public abstract class AssociationRemoteController<T extends Resource<T> & SubResourceContainer, V extends Resource<V>>
        extends SolidifyController {

  @Autowired
  protected CompositeResourceService<T> itemService;

  /**********************************************/

  @GetMapping
  public HttpEntity<Collection<V>> list(@PathVariable String parentid, Pageable pageable) {

    T parentItem = this.itemService.findOne(parentid);
    if (parentItem == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }

    Collection<V> collection = this.itemService.getSubResources(this.getSubResourceIds(parentItem),
            this.getSubResourceType());

    if (collection != null) {
      for (final V v : collection.get_data()) {
        addLinks(parentid, v);
      }
      collection.add(linkTo(this.getClass(), parentid).withSelfRel());
      collection.add((Tool.parentLink((linkTo(this.getClass(), parentid)).toUriComponentsBuilder()))
              .withRel(ActionName.PARENT));
      this.addPageLinks(linkTo(this.getClass(), parentid), collection, pageable);

      return new ResponseEntity<>(collection, HttpStatus.OK);

    } else {
      return new ResponseEntity<>(HttpStatus.GONE);
    }
  }

  @GetMapping(value = SolidifyConstants.URL_ID)
  public HttpEntity<V> get(@PathVariable String parentid, @PathVariable String id) {

    T parentItem = this.itemService.findOne(parentid);
    if (parentItem == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }

    /*
     * Check the requested child Resource is linked to the parent Resource before fetching it (we do not
     * fetch Resources that are not linked to the parent just to be coherent)
     */
    List<String> ids = this.getSubResourceIds(parentItem);
    if (ids != null && ids.contains(id)) {

      V item = this.itemService.getSubResource(id, this.getSubResourceType());

      if (item != null) {
        addLinks(parentid, item);
        return new ResponseEntity<>(item, HttpStatus.OK);
      } else {
        return new ResponseEntity<>(HttpStatus.GONE);
      }
    } else {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
  }

  @PostMapping
  public HttpEntity<List<V>> create(@PathVariable String parentid, @RequestBody String[] ids) {
    boolean noError = true;
    T parentItem = this.itemService.findOne(parentid);
    if (parentItem == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }

    List<V> list = new ArrayList<>();

    for (String id : ids) {

      V childItem = this.itemService.getSubResource(id, this.getSubResourceType());

      if (childItem == null) {
        return new ResponseEntity<>(HttpStatus.NOT_FOUND);
      }

      noError &= parentItem.addItem(childItem);
      list.add(childItem);
    }

    // Save entity if no error
    if (noError) {
      this.setAuthenticatedUserProperties(parentItem);
      this.itemService.save(parentItem);
      return new ResponseEntity<>(list, HttpStatus.CREATED);
    }
    return new ResponseEntity<>(HttpStatus.INTERNAL_SERVER_ERROR);
  }

  @DeleteMapping(value = SolidifyConstants.URL_ID)
  public HttpStatus delete(@PathVariable String parentid, @PathVariable String id) {
    return deleteList(parentid, new String[] { id });
  }

  @DeleteMapping
  public HttpStatus deleteList(@PathVariable String parentid, @RequestBody String[] ids) {
    boolean noError = true;
    T parentItem = this.itemService.findOne(parentid);
    if (parentItem == null) {
      return HttpStatus.NOT_FOUND;
    }

    for (String id : ids) {
      V childItem = this.itemService.getSubResource(id, this.getSubResourceType());

      /*
       * Check the child Resource still exists
       */
      // TODO: if Resource does not exist anymore, an old resId reference can persist in parent item
      // property
      // Should we update RESTResource logic to clear the list of ids instead of passing a
      // RESTResource to removeItem() ?
      if (childItem == null) {
        return HttpStatus.NOT_FOUND;
      }

      /*
       * Check the child Resource is linked to the parent item
       */
      List<String> idsList = this.getSubResourceIds(parentItem);
      if (idsList == null || !idsList.contains(id)) {
        return HttpStatus.GONE;
      }

      /*
       * Remove link from parent Resource to child Resource
       */
      noError &= parentItem.removeItem(childItem);

    }

    if (noError) {
      this.setAuthenticatedUserProperties(parentItem);
      this.itemService.save(parentItem);
      return HttpStatus.OK;
    }

    return HttpStatus.GONE;
  }

  @RequestMapping(value = SolidifyConstants.URL_ID, method = { RequestMethod.PUT,
          RequestMethod.PATCH })
  public HttpEntity<V> update(@PathVariable String parentid, @PathVariable String id,
          @RequestBody T t2) {
    return new ResponseEntity<>(HttpStatus.NOT_IMPLEMENTED);
  }

  protected void addLinks(String parentid, V v) {
    v.removeLinks();
    v.addLinks(linkTo(methodOn(this.getClass()).list(parentid, null)), false, false);
  }

  /**********************************************************/

  protected abstract List<String> getSubResourceIds(T item);

  protected abstract Class<V> getSubResourceType();
}
