package ch.unige.solidify.controller;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.nio.file.Paths;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.InputStreamResource;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.hateoas.RepresentationModel;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.rest.SubResource;
import ch.unige.solidify.rest.SubResourceContainer;
import ch.unige.solidify.rest.Tool;
import ch.unige.solidify.security.NoPermissions;
import ch.unige.solidify.service.CompositeResourceService;
import ch.unige.solidify.util.FileTool;

/**
 * A controller for subresources
 *
 * @param <T>
 *          the type of the main resource
 * @param <V>
 *          the type of the subresource
 */
public abstract class SubResourceController<T extends Resource<T> & SubResourceContainer, V extends Resource<V>>
        extends SolidifyController {

  @Autowired
  protected CompositeResourceService<T> itemService;

  @Autowired
  private CompositeResourceService<V> subResourceService;

  @GetMapping
  public HttpEntity<Collection<V>> list(@PathVariable String parentid, @ModelAttribute V filterItem,
          Pageable pageable) {

    this.checkIfResourceExists(parentid);

    Page<V> list;
    if (filterItem instanceof SubResource) {
      ((SubResource) filterItem).setParentResourceId(parentid);
      list = subResourceService.findAll(filterItem.getSpecification(), pageable);
    } else {
      if (pageable.getSort().isSorted()) {
        throw new UnsupportedOperationException(messageService.get("error.sort.unsupported",
                new Object[] { getMainResourceClass().toString(),
                        getSubResourceClass().toString() }));
      }
      list = itemService
              .findByParentId(parentid, getMainResourceClass(), getSubResourceClass(), pageable);
    }

    for (final V v : list) {
      addLinks(parentid, v);
    }
    Collection<V> collection = processPageAndLinks(list, parentid, pageable);
    return new ResponseEntity<>(collection, HttpStatus.OK);
  }

  protected Collection<V> processPageAndLinks(Page<V> list, String parentid, Pageable pageable) {
    Collection<V> collection = new Collection<>(list, pageable);
    collection.add(linkTo(this.getClass(), parentid).withSelfRel());
    collection.add((Tool.parentLink((linkTo(this.getClass(), parentid)).toUriComponentsBuilder()))
            .withRel(ActionName.PARENT));
    this.addSortLinks(linkTo(this.getClass(), parentid, pageable), collection);
    this.addPageLinks(linkTo(this.getClass(), parentid), collection, pageable);
    this.addOthersLinks(parentid, collection);
    return collection;
  }

  @GetMapping(value = SolidifyConstants.URL_ID)
  public HttpEntity<V> get(@PathVariable String parentid, @PathVariable String id) {
    this.checkIfResourceExists(parentid);
    V item = this.itemService
            .findByParentIdAndChildId(parentid, getMainResourceClass(), getSubResourceClass(), id);
    if (item != null) {
      addLinks(parentid, item);
      return new ResponseEntity<>(item, HttpStatus.OK);
    }
    return new ResponseEntity<>(HttpStatus.NOT_FOUND);
  }

  @NoPermissions
  @GetMapping(value = SolidifyConstants.URL_ID_PLUS_SEP + ActionName.DL)
  public HttpEntity<InputStreamResource> download(@PathVariable String parentid,
          @PathVariable String id) {
    this.checkIfResourceExists(parentid);
    V item = this.itemService
            .findByParentIdAndChildId(parentid, getMainResourceClass(), getSubResourceClass(), id);
    if (item == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    try {
      URI file = item.getDownloadUri();
      InputStream is = new FileInputStream(Paths.get(file).toFile());
      return download(item.getClass().getName(), id, is,
              file.getPath().substring(file.getPath().lastIndexOf('/') + 1),
              FileTool.getContentType(file).toString(), FileTool.getSize(file));
    } catch (IOException e) {
      throw new SolidifyRuntimeException(e.getMessage());
    }
  }

  protected abstract Class<T> getMainResourceClass();

  protected abstract Class<V> getSubResourceClass();

  protected abstract void addLinks(String parentid, V v);

  protected <W extends RepresentationModel<W>> void addOthersLinks(String parentid, W w) {

  }

  protected void checkIfResourceExists(String id) {
    // if resource doen't exist throw a NoSuchElementElement catched by SolidifyControllerAdvice
    this.itemService.findOne(id);
  }
}