package ch.unige.solidify.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.springframework.http.MediaType;
import org.springframework.util.FileSystemUtils;

import ch.unige.solidify.SolidifyConstants;

public class FileTool {
  private static final Logger log = LoggerFactory.getLogger(FileTool.class);

  /**
   * This constructor should not be called
   */
  private FileTool() {
    throw new IllegalStateException(SolidifyConstants.TOOL_CLASS);
  }

  /**
   * Return true if the path corresponds to a folder
   * 
   * @param path
   *          the path to the supposed folder
   * @return true if it's a folder
   */
  public static boolean isFolder(Path path) {
    return path.toFile().isDirectory();
  }

  /**
   * Ensure that folder exists. Create it if not exists.
   * 
   * @param path
   *          the path to the supposed folder
   * @return true folder exists after execution of the method
   */
  public static boolean ensureFolderExists(Path path) {
    path.toFile().mkdirs();
    return checkFile(path) && isFolder(path);
  }

  /**
   * Scan a folder and return the list of all paths satisfying a predicate.
   * 
   * @param folder
   *          The path of the folder to scan.
   * @param predicate
   *          The predicate used to filter the list of paths.
   * @return the list of all paths satisfying the predicate. Return an empty list if an IOException
   *         occurs.
   */
  public static List<Path> scanFolder(Path folder, Predicate<? super Path> predicate) {
    try (Stream<Path> pathStream = Files.list(folder)) {
      return pathStream.filter(predicate).collect(Collectors.toList());
    } catch (IOException e) {
      log.error("IOException when scanning folder " + folder, e);
    }
    return new ArrayList<>();
  }

  /**
   * Scan the classpath and return the list of all paths satisfying a pattern.
   * 
   * @param pattern
   *          The pattern used to filter the list of paths.
   * @return the list of all paths satisfying the pattern. If an IOException occurs stop the scanning
   *         and return the paths already found.
   */
  public static List<Resource> scanClassPath(String pattern) {
    List<Resource> fileList = new ArrayList<>();
    ResourcePatternResolver resolver = new PathMatchingResourcePatternResolver();
    try {
      fileList.addAll(Arrays.asList(resolver.getResources(pattern)));
    } catch (IOException e) {
      log.error("IOException when scanning class path with pattern " + pattern, e);
    }
    return fileList;
  }

  /**
   * Return the list of paths of the regular files in a folder
   * 
   * @param folder
   *          the path of the folder
   * @return the list of regular files Return an empty list if an IOException occurs.
   */
  public static List<Path> readFolder(Path folder) {
    try (Stream<Path> directory = Files.list(folder)) {
      return directory.filter(Files::isRegularFile).collect(Collectors.toList());
    } catch (IOException e) {
      log.error("IOException when reading folder " + folder, e);
    }
    return new ArrayList<>();
  }

  /**
   * Get all regular files paths recursively in a folder
   * 
   * @param dir
   *          the path to the folder
   * @return the list of regular files
   */
  public static List<Path> findFiles(Path dir) {
    try (Stream<Path> filesStream = Files.find(dir, Integer.MAX_VALUE,
            (path, basicFileAttributes) -> basicFileAttributes.isRegularFile())) {
      return filesStream.collect(Collectors.toList());
    } catch (IOException e) {
      log.error("IOException when finding files in " + dir, e);
    }

    return new ArrayList<>();
  }

  /**
   * Delete a folder
   * 
   * @param folder
   *          The path of the folder
   * @return true on success
   */
  public static boolean deleteFolder(Path folder) {
    if (FileSystemUtils.deleteRecursively(folder.toFile())) {
      log.info("Folder " + folder.toString() + " deleted");
      return true;
    }
    log.info("Folder " + folder.toString() + " cannot be deleted");
    return false;
  }

  /**
   * @param uri
   *          the uri corresponding to the path to the file
   * @return the file size as a long type
   */
  public static long getSize(URI uri) {
    return getSize(Paths.get(uri));
  }

  /**
   * @param path
   *          the path to the file
   * @return the file size as a long type
   */
  public static long getSize(Path path) {
    try {
      return Files.size(path);
    } catch (IOException e) {
      log.error("Error while getting size for " + path, e);
      return 0;
    }
  }

  /**
   * @param uri
   *          the uri corresponding to the path to the file
   * @return the media type
   */
  public static MediaType getContentType(URI uri) {
    return getContentType(Paths.get(uri));
  }

  /**
   * @param path
   *          the path to the file
   * @return the media type of the file
   */
  public static MediaType getContentType(Path path) {
    try {
      String fileType = Files.probeContentType(path);
      if (fileType == null) {
        return MediaType.APPLICATION_OCTET_STREAM;
      }
      return MediaType.valueOf(fileType);
    } catch (IOException ioException) {
      log.error("Error while getting content type for " + path, ioException);
      return MediaType.APPLICATION_OCTET_STREAM;
    }
  }

  /**
   * @param uri
   *          the uri corresponding to the path to the file
   * @return true if the it's a normal file
   */
  public static boolean isFile(URI uri) {
    return isFile(Paths.get(uri));
  }

  /**
   * @param path
   *          the path to the file
   * @return true if the it's a normal file
   */
  public static boolean isFile(Path path) {
    return path.toFile().isFile();
  }

  /**
   * @param uri
   *          the uri corresponding to the path to the file
   * @return true if it's a normal file or a directory
   */
  public static boolean checkFile(URI uri) {
    return checkFile(Paths.get(uri));
  }

  /**
   * @param path
   *          the path to the file
   * @return true if it's a normal file or a directory
   */
  public static boolean checkFile(Path path) {
    return path.toFile().exists();
  }

  /**
   * @param src
   *          source path
   * @param dest
   *          destination path
   * @return true if operation succeed
   */
  public static boolean moveFile(Path src, Path dest) {
    if (FileTool.copyFile(src, dest)) {
      return FileTool.deleteFolder(src);
    }
    return false;
  }

  /**
   * Rename a file
   * 
   * @param source
   *          the source URI
   * @param destination
   *          the destination URI
   * @param force
   *          if true replace if destination exists
   * @throws IOException
   *           if an I/O error occurs
   */
  public static void renameFile(URI source, URI destination, boolean force) throws IOException {
    renameFile(Paths.get(source), Paths.get(destination), force);
  }

  /**
   * Rename a file
   * 
   * @param source
   *          the source path
   * @param destination
   *          the destination path
   * @param force
   *          if true replace if destination exists
   * @throws IOException
   *           if an I/O error occurs
   */
  public static void renameFile(Path source, Path destination, boolean force) throws IOException {
    if (force) {
      Files.move(source, destination, StandardCopyOption.REPLACE_EXISTING);
    } else {
      Files.move(source, destination);
    }
  }

  /**
   * Copy a file
   * 
   * @param source
   *          the source URI
   * @param destination
   *          the destination URI
   * @return true if operation succeed
   */
  public static boolean copyFile(URI source, URI destination) {
    return copyFile(Paths.get(source), Paths.get(destination));
  }

  /**
   * Copy a file
   * 
   * @param source
   *          the source path
   * @param destination
   *          the destination path
   * @return true if operation succeed
   */
  public static boolean copyFile(Path source, Path destination) {
    try {
      if (!FileTool.ensureFolderExists(destination.getParent())) {
        log.error("Cannot create directory: " + destination.getParent().toString());
        return false;
      }
      Files.copy(source, destination, StandardCopyOption.COPY_ATTRIBUTES);
    } catch (IOException e) {
      log.error("Error while copying file from " + source + " to " + destination, e);
      return false;
    }
    return true;
  }

  /**
   * Delete a file
   * 
   * @param uri
   *          the URI to the file to delete
   * @return true if operation succeed
   */
  public static boolean deleteFile(URI uri) {
    return deleteFile(Paths.get(uri));
  }

  /**
   * Delete a file
   * 
   * @param path
   *          the path to the file to delete
   * @return true if succeed
   */
  public static boolean deleteFile(Path path) {
    try {
      if (checkFile(path)) {
        Files.delete(path);
        log.info("File " + path.toString() + " deleted");
      }
    } catch (IOException e) {
      log.error("Error while deleting file " + path, e);
      return false;
    }
    return true;
  }

  public static String toString(InputStream inputStream) throws IOException {
    StringBuilder resultStringBuilder = new StringBuilder();
    try (BufferedReader br = new BufferedReader(new InputStreamReader(inputStream))) {
      String line;
      while ((line = br.readLine()) != null) {
        resultStringBuilder.append(line).append("\n");
      }
    }
    return resultStringBuilder.toString();
  }

}
