package ch.unige.solidify.test.util;

import static org.assertj.core.api.Assertions.fail;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.ClassPathResource;
import org.springframework.core.io.FileSystemResource;
import org.springframework.test.context.junit4.SpringRunner;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.exception.SolidifyProcessingException;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.XMLTool;

@RunWith(SpringRunner.class)
public class XMLTest {
  private static final Logger log = LoggerFactory.getLogger(XMLTest.class);

  private static String PREFIX = "test/test";
  private static String DLCM = "dlcm/dlcm";

  private enum XML {
    SCHEMA(0), GOOD(1), GOOD_WO_SCHEMA(2), GOOD_WRONG_SCHEMA(3), NOT_VALID(4), NOT_WELLFORMED(5);
    private final int value;

    private XML(int value) {
      this.value = value;
    }

    public int getValue() {
      return value;
    }
  }

  @Test
  public void transformTest()
          throws SolidifyCheckingException, SolidifyProcessingException, IOException {
    Path xml = this.getPath("dlcm/dlcm_datacite-1.0.xml");
    Path xsl = this.getPath("xslt/datacite2oai_dc.xsl");
    String newXml = xml.getParent().toString() + File.separatorChar
            + xml.getFileName().toString().replaceFirst(".xml", "-out.xml");
    XMLTool.wellformed(xml);
    XMLTool.wellformed(xsl);
    XMLTool.transform(xml, xsl, newXml);
    XMLTool.wellformed(this.getPath(newXml));
    assertTrue(FileTool.deleteFile(Paths.get(newXml)));
  }

  @Test
  public void transformStringTest()
          throws SolidifyCheckingException, SolidifyProcessingException, IOException {
    Path xmlSource = this.getPath("dlcm/dlcm_datacite-1.0.xml");
    Path xslSource = this.getPath("xslt/datacite2oai_dc.xsl");
    String xml = XMLTool.xml2String(xmlSource);
    String xsl = XMLTool.xml2String(xslSource);
    String newXml = xmlSource.getParent().toString() + File.separatorChar
            + xmlSource.getFileName().toString().replaceFirst(".xml", "-out.xml");

    XMLTool.wellformed(xmlSource);
    XMLTool.wellformed(xslSource);
    String xmlResult = XMLTool.transform(xml, xsl);
    try (PrintWriter out = new PrintWriter(newXml)) {
      out.println(xmlResult);
    }
    XMLTool.wellformed(this.getPath(newXml));
    assertTrue(FileTool.deleteFile(Paths.get(newXml)));
  }

  @Test
  public void transformStringWithParamTest()
          throws SolidifyCheckingException, SolidifyProcessingException, IOException {
    Path xmlSource = this.getPath("dlcm/dlcm_datacite-1.0.xml");
    Path xslSource = this.getPath("xslt/datacite2oai_dc_w_param.xsl");
    Map<String, String> params = new HashMap<>();
    params.put("param1", "value1");
    params.put("param2", "value2");
    params.put("param3", "value3");

    String xml = XMLTool.xml2String(xmlSource);
    String xsl = XMLTool.xml2String(xslSource);
    String newXml = xmlSource.getParent().toString() + File.separatorChar
            + xmlSource.getFileName().toString().replaceFirst(".xml", "-out.xml");

    XMLTool.wellformed(xmlSource);
    XMLTool.wellformed(xslSource);
    String xmlResult = XMLTool.transform(xml, xsl, params);
    try (PrintWriter out = new PrintWriter(newXml)) {
      out.println(xmlResult);
    }
    XMLTool.wellformed(this.getPath(newXml));
    assertTrue(FileTool.deleteFile(Paths.get(newXml)));
  }

  @Test
  public void wellFormTest() throws SolidifyCheckingException {
    List<Path> fileList = loadFilesTest(PREFIX);
    XMLTool.wellformed(fileList.get(XML.SCHEMA.getValue()));
    XMLTool.wellformed(fileList.get(XML.GOOD.getValue()));
    XMLTool.wellformed(fileList.get(XML.NOT_VALID.getValue()));
  }

  @Test(expected = SolidifyCheckingException.class)
  public void noWellFormTest() throws SolidifyCheckingException {
    // Temporarily redirect output stream to avoid stack traces in build log
    PrintStream systemOutStream = System.out;
    System.setOut(new java.io.PrintStream(new java.io.OutputStream() {
      @Override
      public void write(int i) {
      }
    }));

    XMLTool.wellformed(loadFilesTest(PREFIX).get(XML.NOT_WELLFORMED.getValue()));
    XMLTool.wellformed(loadFilesTest(DLCM).get(XML.NOT_WELLFORMED.getValue()));

    System.setOut(systemOutStream);
  }

  @Test
  public void validationTest() throws SolidifyCheckingException {
    List<Path> fileList = loadFilesTest(PREFIX);
    List<URL> xsdList = loadSchemaTest(PREFIX);
    XMLTool.validate(xsdList, fileList.get(XML.GOOD.getValue()));
    XMLTool.validate(xsdList, fileList.get(XML.GOOD_WO_SCHEMA.getValue()));
    XMLTool.validate(xsdList, fileList.get(XML.GOOD_WRONG_SCHEMA.getValue()));
  }

  @Test(expected = SolidifyCheckingException.class)
  public void noValidationTest() throws SolidifyCheckingException {
    List<Path> fileList = loadFilesTest(PREFIX);
    List<URL> xsdList = loadSchemaTest(PREFIX);
    XMLTool.validate(xsdList, fileList.get(XML.NOT_VALID.getValue()));
    XMLTool.validate(xsdList, fileList.get(XML.NOT_WELLFORMED.getValue()));
  }

  @Ignore("Depending on external XML schemas")
  public void validationTestStandards() throws IOException, SolidifyCheckingException {
    String[] dirList = { "DataCite-AllIn1-4.0", "DataCite-4.0", "METS-1.11", "PREMIS-3.0" };
    int count = 0;
    for (String dir : dirList) {
      log.info(dir);
      List<Path> xmlList = new ArrayList<>();
      List<URL> xsdList = new ArrayList<>();
      for (Path p : FileTool.readFolder(getPath(dir))) {
        if (p.toString().endsWith(".xml")) {
          xmlList.add(p);
        }
      }
      for (Path p : FileTool.readFolder(getPath(SolidifyConstants.SCHEMA_HOME))) {
        String f = p.getFileName().toString();
        if (f.startsWith(dir) && f.endsWith(".xsd")) {
          log.info("*Schema:" + f);
          xsdList.add(p.toUri().toURL());
        }
      }

      for (Path xml : xmlList) {
        count++;
        log.info("-" + xml.getFileName().toString());
        XMLTool.validate(xsdList, xml);
      }
    }
    log.info(count + " executed tests");
  }

  @Test
  public void wellFormTestDLCM() throws SolidifyCheckingException {
    List<Path> fileList = loadFilesTest(DLCM);
    XMLTool.wellformed(fileList.get(XML.SCHEMA.getValue()));
    XMLTool.wellformed(fileList.get(XML.SCHEMA.getValue()));
    XMLTool.wellformed(fileList.get(XML.GOOD.getValue()));
    XMLTool.wellformed(fileList.get(XML.NOT_VALID.getValue()));
  }

  @Test
  public void wellFormTestDLCMSchemas() throws URISyntaxException, SolidifyCheckingException {
    List<URL> xsdList = loadSchemaTest(DLCM);
    for (URL xsd : xsdList) {
      XMLTool.wellformed(Paths.get(xsd.toURI()));
    }
  }

  // @Test
  public void validationTestDLCM() throws IOException, SolidifyCheckingException {
    List<Path> fileList = loadFilesTest(DLCM);
    List<URL> xsdList = loadSchemaTest(DLCM);

    XMLTool.validate(xsdList, fileList.get(XML.GOOD.getValue()));
    XMLTool.validate(xsdList, getPath("dlcm/dlcm.xml"));
    XMLTool.validate(xsdList, getPath("dlcm/dlcm_datacite.xml"));
    XMLTool.validate(xsdList, fileList.get(XML.GOOD.getValue()));
    XMLTool.validate(xsdList, fileList.get(XML.GOOD_WO_SCHEMA.getValue()));
    XMLTool.validate(xsdList, fileList.get(XML.GOOD_WRONG_SCHEMA.getValue()));
  }

  @Test(expected = SolidifyCheckingException.class)
  public void noValidationTestDLCM() throws IOException, SolidifyCheckingException {
    List<Path> fileList = loadFilesTest(DLCM);
    List<URL> xsdList = loadSchemaTest(DLCM);

    XMLTool.validate(xsdList, fileList.get(XML.NOT_VALID.getValue()));
    XMLTool.validate(xsdList, fileList.get(XML.NOT_WELLFORMED.getValue()));
  }

  private List<Path> loadFilesTest(String prefix) {
    List<Path> list = new ArrayList<>();
    try {
      list.add(getPath(prefix + ".xsd")); // 0 = XML Schema
      list.add(getPath(prefix + "_w_schema.xml")); // 1 = Good XML
      list.add(getPath(prefix + "_wo_schema.xml")); // 2 = Good XML without schema
      list.add(getPath(prefix + "_w_wrong_schema.xml")); // 3 = Good XML without schema
      list.add(getPath(prefix + "_not_valid.xml")); // 4 = Not Valid XML
      list.add(getPath(prefix + "_not_well_formed.xml")); // 5 = Not well-formed XML
    } catch (IOException e) {
      fail(e.getMessage());
    }

    return list;
  }

  private List<URL> loadSchemaTest(String prefix) {
    List<URL> list = new ArrayList<>();
    try {
      list.add(getPath(prefix + ".xsd").toUri().toURL());
    } catch (IOException e) {
      fail(e.getMessage());
    }

    return list;
  }

  private Path getPath(String f) throws IOException {
    try {
      return new ClassPathResource(f).getFile().toPath();
    } catch (IOException e) {
      return new FileSystemResource(f).getFile().toPath();
    }
  }

}