package ch.unige.solidify.test.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.List;

import org.junit.AssumptionViolatedException;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.Stopwatch;
import org.junit.runner.Description;
import org.junit.runner.RunWith;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.io.ClassPathResource;
import org.springframework.test.context.junit4.SpringRunner;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.ZipTool;

@RunWith(SpringRunner.class)
public class ZipTest {
  private static final Logger log = LoggerFactory.getLogger(ZipTest.class);

  @Rule
  public final Stopwatch stopwatch = new Stopwatch() {
    @Override
    protected void succeeded(long nanos, Description description) {
      log.info(description.getMethodName() + " succeeded, duration=" + displayDuration(nanos));
    }

    private String displayDuration(long nanos) {
      double seconds = nanos / 1000000000.0;
      return seconds + "s";
    }

    @Override
    protected void failed(long nanos, Throwable e, Description description) {
      log.info(description.getMethodName() + " failed, duration=" + displayDuration(nanos));
    }

    @Override
    protected void skipped(long nanos, AssumptionViolatedException e, Description description) {
      log.info(description.getMethodName() + " skipped,, duration=" + displayDuration(nanos));
    }

    @Override
    protected void finished(long nanos, Description description) {
      log.info(description.getMethodName() + " finished, duration=" + displayDuration(nanos));
    }

  };

  private final static String ZIP_DIR = "dlcm";
  private final static String INCLUDE = "include";
  private final static String ZIP_ARCHIVE = "test.zip";
  private final static int ZIP_FILE_NUMBER = 17;
  private final static int ZIP_INCLUDE_NUMBER = 8;
  private final static int ZIP_XSD_NUMBER = 9;
  private final static String UNZIP_DIR = "unzip";

  @Test
  public void zipArchiveTest() throws IOException {
    Path zipDir = Paths.get(new ClassPathResource(ZIP_DIR).getURI());
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    assertNotNull(zipDir);
    assertTrue(zipTool.zipFiles(zipDir));
    assertTrue(zipTool.checkZip());
    assertEquals(ZIP_FILE_NUMBER, zipTool.getEntryNumber());
    assertTrue(FileTool.deleteFile(Paths.get(ZIP_ARCHIVE)));
  }

  @Test
  public void unzipFolderTest() throws IOException {
    Path zipDir = Paths.get(new ClassPathResource(ZIP_DIR).getURI());
    Path unzipDir = Paths
            .get(System.getProperty("java.io.tmpdir") + File.separatorChar + UNZIP_DIR);
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    assertTrue(zipTool.zipFiles(zipDir.toUri()));
    assertTrue(zipTool.unzipFolder(unzipDir, INCLUDE + SolidifyConstants.URL_SEP));
    List<Path> fileList = FileTool.readFolder(unzipDir);
    assertEquals("Wrong file number", ZIP_INCLUDE_NUMBER, fileList.size());
    assertTrue("Cannot delete files", FileTool.deleteFile(Paths.get(ZIP_ARCHIVE)));
    assertTrue("Cannot delete folder", FileTool.deleteFolder(unzipDir));
  }

  @Test
  public void unzipOSDependantFolderTest() throws IOException {
    Path zipDir = Paths.get(new ClassPathResource(ZIP_DIR).getURI());
    Path unzipDir = Paths
            .get(System.getProperty("java.io.tmpdir") + File.separatorChar + UNZIP_DIR);
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    assertTrue(zipTool.zipFiles(zipDir.toUri()));
    assertTrue(zipTool.unzipFolder(unzipDir, INCLUDE + File.separator));
    List<Path> fileList = FileTool.readFolder(unzipDir);
    assertEquals("Wrong file number", ZIP_INCLUDE_NUMBER, fileList.size());
    assertTrue("Cannot delete files", FileTool.deleteFile(Paths.get(ZIP_ARCHIVE)));
    assertTrue("Cannot delete folder", FileTool.deleteFolder(unzipDir));
  }

  @Test
  public void unzipArchiveTest() throws IOException {
    Path zipDir = Paths.get(new ClassPathResource(ZIP_DIR).getURI());
    Path unzipDir = Paths
            .get(System.getProperty("java.io.tmpdir") + File.separatorChar + UNZIP_DIR);
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    assertTrue(zipTool.zipFiles(zipDir.toUri()));
    assertTrue(zipTool.unzipFiles(unzipDir));
    List<Path> fileList = FileTool.readFolder(unzipDir);
    fileList.addAll(FileTool.readFolder(unzipDir.resolve(INCLUDE)));
    assertEquals(ZIP_FILE_NUMBER, fileList.size());
    assertTrue(FileTool.deleteFile(Paths.get(ZIP_ARCHIVE)));
    assertTrue(FileTool.deleteFolder(unzipDir));
  }

  @Test
  public void partialUnzipArchiveTest() throws IOException {
    Path zipDir = Paths.get(new ClassPathResource(ZIP_DIR).getURI());
    Path unzipDir = Paths
            .get(System.getProperty("java.io.tmpdir") + File.separatorChar + UNZIP_DIR);
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    List<String> extractList = Arrays.asList("dlcm.xsd",
            "include" + SolidifyConstants.URL_SEP + "datacite-");
    assertTrue(zipTool.zipFiles(zipDir));
    assertTrue(zipTool.unzipFiles(unzipDir, extractList));
    List<Path> fileList = FileTool.readFolder(unzipDir);
    fileList.addAll(FileTool.readFolder(unzipDir.resolve(INCLUDE)));
    assertEquals(ZIP_XSD_NUMBER, fileList.size());
    assertTrue(FileTool.deleteFile(Paths.get(ZIP_ARCHIVE)));
    assertTrue(FileTool.deleteFolder(unzipDir));
  }

  @Test
  public void partialUnzipArchiveTest2() throws IOException {
    Path zipDir = Paths.get(new ClassPathResource(ZIP_DIR).getURI());
    Path unzipDir = Paths
            .get(System.getProperty("java.io.tmpdir") + File.separatorChar + UNZIP_DIR);
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    List<String> extractList = Arrays.asList("dlcm.xsd",
            "include" + SolidifyConstants.URL_SEP + "datacite-");
    assertTrue(zipTool.zipFiles(zipDir));
    assertTrue(zipTool.unzipFiles(unzipDir, extractList));
    List<Path> fileList = FileTool.readFolder(unzipDir);
    fileList.addAll(FileTool.readFolder(unzipDir.resolve(INCLUDE)));
    assertEquals(ZIP_XSD_NUMBER, fileList.size());
    assertTrue(FileTool.deleteFile(Paths.get(ZIP_ARCHIVE)));
    assertTrue(FileTool.deleteFolder(unzipDir));
  }

  @Test
  public void partialUnzipArchiveTest2bis() throws IOException {
    Path zipDir = Paths.get(new ClassPathResource(ZIP_DIR).getURI());
    Path unzipDir = Paths
            .get(System.getProperty("java.io.tmpdir") + File.separatorChar + UNZIP_DIR);
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    List<String> extractList = Arrays.asList("dlcm.xsd",
            "include" + SolidifyConstants.URL_SEP + "datacite-");
    assertTrue(zipTool.zipFiles(zipDir));
    assertTrue(zipTool.unzipFiles(unzipDir, extractList));
    List<Path> fileList = FileTool.readFolder(unzipDir);
    fileList.addAll(FileTool.readFolder(unzipDir.resolve(INCLUDE)));
    assertEquals(ZIP_XSD_NUMBER, fileList.size());
    assertTrue(FileTool.deleteFile(Paths.get(ZIP_ARCHIVE)));
    assertTrue(FileTool.deleteFolder(unzipDir));
  }

  @Test
  public void unzipArchiveTest1() throws IOException {
    Path zipDir = Paths.get(new ClassPathResource(ZIP_DIR).getURI());
    Path unzipDir = Paths
            .get(System.getProperty("java.io.tmpdir") + File.separatorChar + UNZIP_DIR);
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    assertTrue(zipTool.zipFiles(zipDir));
    assertTrue(zipTool.unzipFiles(unzipDir));
    assertTrue(FileTool.deleteFile(Paths.get(ZIP_ARCHIVE)));
    assertTrue(FileTool.deleteFolder(unzipDir));
  }

  @Test
  public void unzipArchiveTest2() throws IOException {
    Path zipDir = Paths.get(new ClassPathResource(ZIP_DIR).getURI());
    Path unzipDir = Paths
            .get(System.getProperty("java.io.tmpdir") + File.separatorChar + UNZIP_DIR);
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    assertTrue(zipTool.zipFiles(zipDir));
    assertTrue(zipTool.unzipFiles(unzipDir, null));
    assertTrue(FileTool.deleteFile(Paths.get(ZIP_ARCHIVE)));
    assertTrue(FileTool.deleteFolder(unzipDir));
  }

}
